package inf101.v18.rogue101.enemies;

import inf101.v18.grid.GridDirection;
import inf101.v18.grid.Location;
import inf101.v18.rogue101.game.Game;
import inf101.v18.rogue101.items.buff.BuffItem;
import inf101.v18.rogue101.items.weapon.BasicBow;
import inf101.v18.rogue101.items.weapon.BasicSword;
import inf101.v18.rogue101.items.weapon.FireStaff;
import inf101.v18.rogue101.items.weapon.MagicWeapon;
import inf101.v18.rogue101.items.weapon.MeleeWeapon;
import inf101.v18.rogue101.items.weapon.RangedWeapon;
import inf101.v18.rogue101.object.Item;
import inf101.v18.rogue101.state.AttackType;
import inf101.v18.rogue101.state.LifeStage;
import inf101.v18.rogue101.state.Occupation;
import inf101.v18.rogue101.state.Personality;
import inf101.v18.util.NPCHelper;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;

/**
 * An enemy NPC
 */
public class Girl extends Enemy {

    private final String name = randomName();
    private final LifeStage lifeStage = LifeStage.getRandom();
    private final Personality personality = Personality.getRandom();
    private final Occupation occupation = Occupation.getRandom();
    private int maxHp;
    private int attack;
    private int defence;
    private int visibility;
    private int damage;
    private List<GridDirection> damageDirections;
    private static final Random random = new Random();
    private List<Class<?>> validItems;
    private static final String[] nameList = {"Milk", "Salad", "Mikan", "Rima", "Suki", "Miki", "Hinata", "Haruna",
            "Asuna"};

    /**
     * Instantiates a new Girl enemy NPC
     */
    public Girl() {
        setStats();
        setValidItems();
    }

    private void setStats() {
        // Set stats based on the NPC's age
        switch (lifeStage) {
            case TODDLER:
                maxHp = 30;
                attack = 10;
                defence = 50;
                damage = 1 + random.nextInt(5);
                visibility = 1;
                break;
            case CHILD:
                maxHp = 50;
                attack = 20;
                defence = 40;
                damage = 2 + random.nextInt(5);
                visibility = 2;
                break;
            case TEEN:
                maxHp = 70;
                attack = 25;
                defence = 30;
                damage = 3 + random.nextInt(5);
                visibility = 3;
                break;
            case ADULT:
                maxHp = 100;
                attack = 30;
                defence = 20;
                damage = 4 + random.nextInt(5);
                visibility = 4;
                break;
            case ELDER:
                maxHp = 70;
                attack = 15;
                defence = 35;
                damage = 3 + random.nextInt(5);
                visibility = 2;
                break;
        }
        // Improve stats based on occupation
        if (occupation == Occupation.KNIGHT) {
            attack += 10; //Knights are quite powerful.
        } else if (occupation == Occupation.MAGE) {
            attack += 5; // Mages have lesser range than archers, but more damage.
        }
        // Add random stat improvements
        maxHp += (int) (random.nextGaussian() * 10);
        hp = maxHp;
        attack += (int) (random.nextGaussian() * 5);
        defence += (int) (random.nextGaussian() * 5);
    }

    /**
     * Has a chance of giving a weapon to a girl.
     *
     * @param lvl The current map level
     */
    public void giveWeapon(int lvl) {
        // Randomly grant a weapon based on occupation.
        // The chance of being granted a weapon increases for each level 
        if (random.nextInt(5) >= lvl) {
            return;
        }
        switch (occupation) {
            case KNIGHT:
                backpack.add(new BasicSword());
                break;
            case MAGE:
                backpack.add(new FireStaff());
                break;
            case ARCHER:
                backpack.add(new BasicBow());
        }
    }

    /**
     * Specified which items the current Girl should be able to pick up.
     */
    private void setValidItems() {
        validItems = new ArrayList<>();
        switch (occupation) {
            case ARCHER:
                validItems.add(RangedWeapon.class);
                break;
            case MAGE:
                validItems.add(MagicWeapon.class);
                break;
            case KNIGHT:
                validItems.add(MeleeWeapon.class);
        }
        validItems.add(BuffItem.class);
    }

    /**
     * Picks a "random" name
     *
     * @return A random name
     */
    private String randomName() {
        return nameList[random.nextInt(nameList.length)] + "-chan";
    }

    @Override
    public void doTurn(Game game) {
        currentLocation = game.getLocation();
        List<GridDirection> damageDirections = this.damageDirections;
        this.damageDirections = null;

        // If this NPC has space, pick up any items on the ground
        if (backpack.hasSpace()) {
            Item item = NPCHelper.pickUp(game, validItems);
            if (item != null) {
                backpack.add(item);
                return;
            }
            if (NPCHelper.trackItem(game, validItems)) {
                return;
            }
        }

        // TODO: If attacked on the last turn, flee from the attacker if not within sight range
        if ((personality == Personality.AFRAID || hp < 0.1 * maxHp)) {
            // Flee from the nearest enemy
            if (NPCHelper.flee(game)) {
                return;
            }

            // Flee in the direction this NPC was damaged from. This allows the NPC to flee from ranged attacks outside
            // visible range.
            if (damageDirections != null) {
                for (GridDirection direction : damageDirections) {
                    if (game.canGo(direction)) {
                        game.move(direction);
                        return;
                    }
                }
            }
        }

        // Attack if necessary
        if (willAttack(game) && attack(game)) {
            return;
        }

        // Move a random direction if possible
        List<GridDirection> possibleMoves = game.getPossibleMoves();
        if (!possibleMoves.isEmpty()) {
            Collections.shuffle(possibleMoves);
            game.move(possibleMoves.get(0));
        }
    }

    /**
     * Tries to attack with an attack specified by the Girl's occupation
     *
     * @param game An IGame object
     * @return True if an attack occurred. False otherwise
     */
    private boolean attack(Game game) {
        switch (occupation) {
            case KNIGHT:
                if (NPCHelper.tryAttack(game, 1, AttackType.MELEE)) {
                    return true;
                }
                break;
            case MAGE:
                if (NPCHelper.tryAttack(game, getVision() / 2 + getVision() % 2 == 0 ? 0 : 1, AttackType.MAGIC)) {
                    return true;
                }
            case ARCHER:
                if (NPCHelper.tryAttack(game, getVision(), AttackType.RANGED)) {
                    return true;
                }
        }
        return false;
    }

    /**
     * Checks if the current girl will try to attack.
     *
     * @param game An IGame object
     * @return True if the girl will attack. False otherwise
     */
    private boolean willAttack(Game game) {
        boolean attack = false;
        switch (personality) {
            case CALM:
                // Attack if previously attacked
                if (hp < maxHp) {
                    attack = true;
                }
                break;
            case AFRAID:
                // Attack if cornered
                if (game.getPossibleMoves().isEmpty()) {
                    attack = true;
                }
                break;
            case AGGRESSIVE:
                // Always attack
                attack = true;
                break;
        }
        return attack;
    }

    @Override
    public int getAttack() {
        return attack;
    }

    @Override
    public int getDamage() {
        return damage;
    }

    @Override
    public int getCurrentHealth() {
        return hp;
    }

    @Override
    public int getDefense() {
        return defence;
    }

    @Override
    public int getMaxHealth() {
        return maxHp;
    }

    @Override
    public String getName() {
        return name;
    }

    @Override
    public int getSize() {
        return 30;
    }

    @Override
    public String getPrintSymbol() {
        return "\u001b[95m" + "\uD83D\uDEB6" + "\u001b[0m";
    }

    @Override
    public String getSymbol() {
        return "G";
    }

    @Override
    public int getVision() {
        return visibility;
    }

    @Override
    public int handleDamage(Game game, Item source, int amount) {
        super.handleDamage(game, source, amount);
        Location location = game.getMap().getLocation(source);
        Location location2 = game.getMap().getLocation(this);
        damageDirections = game.locationDirection(location, location2);
        return amount;
    }

}
