package inf101.v18.util;

import inf101.v18.gfx.textmode.BlocksAndBoxes;
import inf101.v18.grid.GridDirection;
import inf101.v18.grid.Location;
import inf101.v18.rogue101.game.Game;
import inf101.v18.rogue101.items.weapon.MagicWeapon;
import inf101.v18.rogue101.items.weapon.MeleeWeapon;
import inf101.v18.rogue101.items.weapon.RangedWeapon;
import inf101.v18.rogue101.items.weapon.Weapon;
import inf101.v18.rogue101.map.MapView;
import inf101.v18.rogue101.object.Actor;
import inf101.v18.rogue101.object.Item;
import inf101.v18.rogue101.object.PlayerCharacter;
import inf101.v18.rogue101.state.AttackType;

import java.util.List;

/**
 * A holder class for methods used by IActors.
 */
public class NPCHelper {

    private NPCHelper() {
    }

    /**
     * Performs a ranged attack on the closest actor (if any) observable by the current actor if range is > 1.
     * Performs a melee attack on an actor (if any) on a neighbouring tile if range == 1.
     * Moves closer to an enemy if an attack is not possible.
     *
     * @param game  An IGame object
     * @param range The range of the equipped weapon
     * @return True if an attack or a move towards an enemy was successful. False otherwise
     */
    public static boolean tryAttack(Game game, int range, AttackType type) {
        List<Location> neighbours = game.getVisible();
        for (Location neighbour : neighbours) {
            MapView map = game.getMap();
            if (map.hasActors(neighbour) && map.getActors(neighbour).get(0) instanceof PlayerCharacter) {
                Location current = game.getLocation();
                List<GridDirection> dirs = game.locationDirection(current, neighbour);
                Actor actor = game.getMap().getActors(neighbour).get(0); //We assume there is only one actor.
                for (GridDirection dir : dirs) {
                    if (type == AttackType.MELEE
                            && current.gridDistanceTo(neighbour) <= 1
                            && current.canGo(dir)
                            && game.getMap().has(current.go(dir), actor)
                    ) {
                        game.attack(dir, actor);
                        return true;
                    }
                    if (range > 1 && current.gridDistanceTo(neighbour) <= range) {
                        game.rangedAttack(dir, actor, type);
                        return true;
                    }
                }
                for (GridDirection dir : dirs) {
                    if (game.canGo(dir)) {
                        game.move(dir);
                        return true;
                    }
                }
            }
        }
        return false;
    }

    /**
     * Tries to find and reach an item of a specified class.
     *
     * @param game       An IGame object
     * @param validItems A list of classes the NPC should look for
     * @return True if an appropriate item was found in the range of the current actor. False otherwise.
     */
    public static boolean trackItem(Game game, List<Class<?>> validItems) {
        List<Location> neighbours = game.getVisible();
        for (Location neighbour : neighbours) {
            for (Item item : game.getMap().getAll(neighbour)) {
                if (!isValid(item, validItems)) {
                    continue;
                }

                // If an item is found, go towards the item
                Location current = game.getLocation();
                List<GridDirection> directions = game.locationDirection(current, neighbour);
                for (GridDirection direction : directions) {
                    if (game.canGo(direction)) {
                        game.move(direction);
                        return true;
                    }
                }
            }
        }
        return false;
    }

    /**
     * Checks whether an item's class is in the list of valid item classes
     *
     * @param item       <p>The item to check</p>
     * @param validItems <p>Valid item classes to look for</p>
     * @return <p>True if the item is valid</p>
     */
    private static boolean isValid(Item item, List<Class<?>> validItems) {
        for (Class<?> validItem : validItems) {
            if (validItem.isInstance(item)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Picks up an item of one of the specified class types.
     *
     * @param game       An IGame object
     * @param validItems A list of classes the NPC should accept
     * @return An item if it was successfully picked up. Null otherwise
     */
    public static Item pickUp(Game game, List<Class<?>> validItems) {
        for (Item item : game.getLocalItems()) {
            for (Class<?> validItem : validItems) {
                if (validItem.isInstance(item)) {
                    return game.pickUp(item);
                }
            }
        }
        return null;
    }

    /**
     * Makes an Actor go to the opposite direction of the closest enemy.
     *
     * <p>Will return false if there are no visible enemies.</p>
     *
     * @param game <p>A Game object</p>
     * @return <p>True if the Actor fled. False otherwise</p>
     */
    public static boolean flee(Game game) {
        List<Location> neighbours = game.getVisible();
        for (Location neighbour : neighbours) {
            if (!game.getMap().hasActors(neighbour)) {
                continue;
            }

            Location current = game.getLocation();
            List<GridDirection> directions = reverseDirections(game.locationDirection(current, neighbour));
            for (GridDirection direction : directions) {
                if (game.canGo(direction)) {
                    game.move(direction);
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Gets the reverse of a given direction
     *
     * @param direction <p>The direction to get the reverse of</p>
     * @return <p>The reverse direction</p>
     */
    public static GridDirection getReverseDirection(GridDirection direction) {
        switch (direction) {
            case SOUTH:
                return GridDirection.NORTH;
            case NORTH:
                return GridDirection.SOUTH;
            case WEST:
                return GridDirection.EAST;
            case EAST:
                return GridDirection.WEST;
            case NORTHEAST:
                return GridDirection.SOUTHWEST;
            case NORTHWEST:
                return GridDirection.SOUTHEAST;
            case SOUTHEAST:
                return GridDirection.NORTHWEST;
            case SOUTHWEST:
                return GridDirection.NORTHEAST;
            case CENTER:
                return GridDirection.CENTER;
            default:
                throw new IllegalArgumentException("Unknown direction encountered");
        }
    }

    /**
     * Reverses a list of directions.
     *
     * @param directions A list directions
     * @return A list of the opposite directions
     */
    private static List<GridDirection> reverseDirections(List<GridDirection> directions) {
        directions.replaceAll(NPCHelper::getReverseDirection);
        return directions;
    }

    /**
     * Generates a hp bar based on item's current hp
     *
     * @return HP bar as string
     */
    public static String hpBar(Item item) {
        StringBuilder str = new StringBuilder(BlocksAndBoxes.BLOCK_HALF);
        int hp = item.getCurrentHealth();
        int hpLen = (int) ((float) hp / item.getMaxHealth() * 10.0);
        if (hp > 0 && hp < 100) {
            str.append("\u001b[91m" + BlocksAndBoxes.BLOCK_BOTTOM + "\u001b[0m");
            hpLen++;
        } else {
            for (int i = 0; i < hpLen; i++) {
                str.append("\u001b[91m" + BlocksAndBoxes.BLOCK_HALF + "\u001b[0m");
            }
        }
        for (int i = 0; i < 10 - hpLen; i++) {
            str.append(" ");
        }
        str.append(BlocksAndBoxes.BLOCK_HALF);
        return str.toString();
    }

    /**
     * Gets the weapon appropriate for the used attack.
     *
     * @param type The attack type
     * @return An IWeapon or null
     */
    public static Weapon getWeapon(AttackType type, Actor actor) {
        Weapon weapon = null;
        switch (type) {
            case MELEE:
                weapon = (Weapon) actor.getItem(MeleeWeapon.class);
                break;
            case RANGED:
                weapon = (Weapon) actor.getItem(RangedWeapon.class);
                break;
            case MAGIC:
                weapon = (Weapon) actor.getItem(MagicWeapon.class);
        }
        return weapon;
    }
}
